import base64


def base64url_decode(input_string: str) -> str:
    """Encode a base64url string to string

    Args:
    -----
        input_string: Input String

    Returns:
    --------
        Decoded string

    """
    input_string = input_string.replace("-", "+").replace("_", "/")
    padding = len(input_string) % 4
    if padding:
        input_string += "=" * (4 - padding)
    decoded_bytes = base64.b64decode(input_string)
    decoded_string = decoded_bytes.decode("utf-8")  # Adjust encoding if needed
    return decoded_string



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Union

import html2text
from langchain.schema import Document

from inarrator.utils import base64url_decode


@dataclass
class IMessage(ABC):
    to: str
    fro: str
    subject: str
    body: str
    email_provider: str = ""

    @classmethod
    @abstractmethod
    def parse_message(cls, **kwargs: Any) -> Union[IMessage, None]:
        """Parse the Raw Message payload and create a message instance.

        Args:
        ----
            kwargs: Message Payload Keyword Argument

        Returns:
        -------
            IMessage Instance

        """
        pass

    def message_to_document(email: IMessage) -> Document:
        """
        Converts an email message into a Document.

        Args:
            email (IMessage): The email message to convert.

        Returns:
            Document: A Langchain Document instance.
        """
        return Document(page_content=str(email))

    def __str__(self) -> str:
        """Returns a string representation of the message."""
        return (
            f"FROM: {self.fro} \n TO: {self.to} \n SUBJECT: {self.subject} \n BODY:\n {self.body}"
        )


@dataclass
class GmailMessage(IMessage):
    to: str
    fro: str
    subject: str
    body: str
    email_provider: str = "Gmail"

    @classmethod
    def parse_message(cls, **kwargs: Any) -> Union[IMessage, None]:
        h = html2text.HTML2Text()
        h.ignore_links = True
        h.ignore_images = True
        raw_payload = kwargs.get("raw_message_payload", {})
        if not raw_payload:
            raise ValueError(
                "In order parse Gmail Email Payload you need to provide raw_message_payload"
            )
        payload = raw_payload.get("payload")
        headers = payload.get("headers")
        body = payload.get("body").get("data")
        if not body:
            return None
        headers_dict = {"To": "", "From": "", "Subject": ""}
        for header in headers:
            if header.get("name") == "From":
                headers_dict["From"] = header.get("value")
            elif header.get("name") == "To":
                headers_dict["To"] = header.get("value")
            elif header.get("Subject") == "Subject":
                headers_dict["Subject"] = header.get("value")

        return cls(
            to=headers_dict.get("To", ""),
            fro=headers_dict.get("From", ""),
            subject=headers_dict.get("Subject", ""),
            body=h.handle(base64url_decode(body)),
        )


@dataclass
class OutlookMessage(IMessage):
    to: str
    fro: str
    subject: str
    body: str
    email_provider: str = "Outlook"

    @classmethod
    def parse_message(cls, **kwargs: Any) -> Union[IMessage, None]:
        h = html2text.HTML2Text()
        h.ignore_links = True
        h.ignore_images = True
        raw_payload = kwargs.get("raw_message_payload", {})
        if not raw_payload:
            raise ValueError(
                "In order parse Outlook Email Payload you need to provide raw_message_payload"
            )
        subject = raw_payload["subject"]
        message_body = h.handle(raw_payload["body"].get("content"))
        fro = raw_payload.get("from").get("emailAddress").get("address")
        to = ";".join(
            [email.get("emailAddress").get("address") for email in raw_payload.get("toRecipients")]
        )
        return cls(to=to, fro=fro, subject=subject, body=message_body)



import concurrent.futures
from abc import ABC, abstractmethod
from concurrent.futures import ProcessPoolExecutor
from typing import Any, Dict, List, Sequence, Tuple

from inarrator.email.email import IEmail
from inarrator.email.message import IMessage
from inarrator.exceptions import AuthenticationError


class IUser(ABC):
    def __init__(self, emails: List[IEmail]):
        self.emails = emails

    @abstractmethod
    def authenticate(self) -> None:
        """Authentication of each email API for the given user."""
        pass

    @classmethod
    @abstractmethod
    def read_email_specific_client(
        cls, email: IEmail, **kwargs: Any
    ) -> Tuple[str, Sequence[IMessage]]:
        """Reads  the user email for the specific email client.

        Args:
        ----
            email: IEmail Instance.

            kwargs: Keyword Argument be used when fetching the email for the given client

        Returns:
        -------
            A tuple consist of IEmail child class name and sequence of fetched messages.

        """

        pass

    @abstractmethod
    def read_latest_emails(self, **kwargs: Any) -> Dict[str, Sequence[IMessage]]:
        """Getting the latest unread emails for a user in a parallel manner.
        Spawns multiple processors to fetch emails from each of the email client for
        the given user in a parallel manner.

        Args:
        ----
            kwargs: Keyword Argument be used when fetching the emails.

        Returns:
        -------
            A dictionary which consist of emails provider and messages from each email provider
        """
        pass

    @abstractmethod
    def summarize_latest_emails(self) -> str:
        pass


class User(IUser):
    def __init__(self, emails: List[IEmail]):
        self.emails = emails

    def authenticate(self) -> None:
        for email in self.emails:
            if not email.service:
                raise AuthenticationError(
                    f"""Authentication Error.
                    Make sure following {email._type} API has not been authenticated
                    """
                )

    @classmethod
    def read_email_specific_client(
        cls, email: IEmail, **kwargs: Any
    ) -> Tuple[str, Sequence[IMessage]]:
        return email._type, email.get_latest_emails(**kwargs)

    def read_latest_emails(self, **kwargs: Any) -> Dict[str, Sequence[IMessage]]:
        emails_dict = {}
        with ProcessPoolExecutor() as executor:
            email_futures = [
                executor.submit(User.read_email_specific_client, email, **kwargs)
                for email in self.emails
            ]
        for future in concurrent.futures.as_completed(email_futures):
            email_type, latest_emails = future.result()
            emails_dict[email_type] = latest_emails

        return emails_dict

    def summarize_latest_emails(self) -> str:
        return ""



class AuthenticationError(Exception):
    """Custom exception class for authentication errors."""

    def __init__(self, message: str = "Authentication failed"):
        self.message = message
        super().__init__(self.message)



"""PROMPTS FOR OUR EMAIL-SUMMARIZER"""
HUGGING_FACE_EMAIL_PROMPT = """Your main task is to summarize an email content.
Please make sure to include the main email content and summarize the main
points. Summarize the following Email Content:

FROM: {fro}
TO: {to}
SUBJECT: {subject}
BODY: {body}

Use the following format:
FROM: This is the person who send the email: {fro}
Summarized-Email-Conten: Email-Summarized Content
"""
STUFF_EMAIL_PROMPT = """
Please summarize the following emails:

{emails}

Summary:
"""



from __future__ import annotations

from typing import List, Union

from langchain.chains import LLMChain
from langchain.chains.summarize import load_summarize_chain
from langchain.chat_models import ChatOpenAI
from langchain.prompts import PromptTemplate

from inarrator.email.message import IMessage
from inarrator.summarizer.base import ISummarizer
from inarrator.summarizer.prompt import STUFF_EMAIL_PROMPT


class GPTModel(ISummarizer):
    def __init__(self, api_token: str, model_name: str):
        self.chain = self.load_chain(api_token, model_name)

    @classmethod
    def load_chain(self, api_token: str, model_name: str) -> LLMChain:
        llm = ChatOpenAI(openai_api_key=api_token, model_name=model_name, temperature=0)
        prompt = PromptTemplate.from_template(STUFF_EMAIL_PROMPT)
        return load_summarize_chain(
            llm, prompt=prompt, chain_type="stuff", document_variable_name="emails"
        )

    def summarize(self, messages: Union[IMessage, List[IMessage]]) -> str:
        if isinstance(messages, IMessage):
            messages = [messages]
        messages = list(map(lambda x: x.message_to_document(), messages))
        return self.chain.run(messages)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Union

from langchain.chains import LLMChain

from inarrator.email.message import IMessage


class ISummarizer(ABC):
    def __init__(self, api_token: str, model_name: str):
        """
        Abstract base class for summarizers.

        Args:
            api_token (str): The API token for the summarization model.
            model_name (str): The name of the summarization model.
        """
        self.chain = self.load_chain(api_token, model_name)

    @classmethod
    @abstractmethod
    def load_chain(self, api_token: str, model_name: str) -> LLMChain:
        """Load a language model chain for summarization.

        Args:
            api_token (str): The API token for the summarization model.
            model_name (str): The name of the summarization model.

        Returns:
            LLMChain: A language model chain for summarization.
        """
        pass

    @abstractmethod
    def summarize(self, messages: Union[IMessage, List[IMessage]]) -> str:
        """Summarize one or more email messages using the loaded language model chain.

        Args:
            message (Union[IMessage, List[IMessage]]): An email message or a list of email messages
            to be summarized.

        Returns:
            str: The summarized content of the message.
        """
        pass



"""Module to Get Emails Data using respective API(s)."""

from __future__ import annotations

import datetime
import json
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, Sequence

import msal
import requests_oauthlib
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from inarrator.email.message import GmailMessage, IMessage, OutlookMessage


class IEmail(ABC):
    """Abstract Class for Email Instances."""

    service: Any = None
    """Main API Python Client"""

    @property
    def _type(self) -> str:
        """Type of Email API."""
        return "IEmail"

    @abstractmethod
    def authenticate(self, **kwargs: Any) -> None:
        """Authenticate the underlying EMAIL API to access user emails.

        Args:
        ----
            kwargs: Key word argument needed to authenticate email API.
        """
        pass

    @abstractmethod
    def get_latest_emails(self, **kwargs: Any) -> Sequence[IMessage]:
        """Get the latest unread emails from user inbox.

        Args:
        ----
            kwargs: Keyword argument needed to filter out the latest emails.

        Returns:
        -------
            List of Email Messages
        """
        pass

    @abstractmethod
    def get_email(self, **kwargs: Any) -> IMessage:
        """Get specific email from user inbox given a unique identifier for the email.

        Args:
        ----
            kwargs: Keyword Argument for the unique id.

        Returns:
        -------
            Email Message
        """
        pass


class Gmail(IEmail):
    service: Any = None

    @property
    def _type(self) -> str:
        return "Gmail"

    def authenticate(self, **kwargs: Any) -> None:
        if "credentials_path" not in kwargs:
            raise ValueError("Need to provide Gmail API credential json file path")
        creds = None
        if os.path.exists("gmail_token.json"):
            creds = Credentials.from_authorized_user_file(
                "gmail_token.json", kwargs.get("gmail_scope")
            )
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    kwargs.get("credentials_path"), kwargs.get("gmail_scope")
                )
                creds = flow.run_local_server(port=0)
            # Save the credentials for the next run
            with open("gmail_token.json", "w") as token:
                token.write(creds.to_json())
        try:
            # Call the Gmail API
            self.service = build("gmail", "v1", credentials=creds)
            self.service.users().labels().list(userId="me").execute()
        except HttpError as error:
            print(f"An error occurred: {error}")

    def get_latest_emails(self, **kwargs: Any) -> Sequence[IMessage]:
        if not self.service:
            raise ValueError("You need to authenticate Gmail API")
        results = (
            self.service.users()
            .messages()
            .list(
                userId="me",
                includeSpamTrash=False,
                maxResults=kwargs.get("gmail_max_emails", 10),
                labelIds=["INBOX"],
                q=kwargs.get("gmail_filters", "is:unread"),
            )
            .execute()
        )
        messages = results.get("messages", [])
        messages_list = []
        # Now get individual messages payload and format them
        for message_dict in messages:
            raw_message_payload = self.get_email(user_email="me", gmail_id=message_dict.get("id"))
            message = GmailMessage.parse_message(raw_message_payload=raw_message_payload)
            if message:
                messages_list.append(message)
        return messages_list

    def get_email(self, **kwargs: Any) -> Any:
        return self.service.users().messages().get(userId="me", id=kwargs.get("gmail_id")).execute()


class OutLook(IEmail):
    service: Any = None

    @property
    def _type(self) -> str:
        return "Outlook"

    def _token(self, **kwargs: Any) -> Dict:
        """Gets the user access token needed to authenticate Microsoft Graph API"""
        if "credentials_path" not in kwargs:
            raise ValueError("Need to provide Outlook API credential json file path")
        with open(kwargs["credentials_path"], "r") as f:
            credentials_json = json.load(f)
            f.close()
        client_instance = msal.PublicClientApplication(
            client_id=credentials_json.get("application_id"), authority=kwargs.get("authority_url")
        )
        result = client_instance.acquire_token_interactive(scopes=kwargs.get("outlook_scope"))
        if "access_token" not in result:
            raise ValueError("Authentication Error for Outlook Client")
        return result

    def _token_expired(self, token_id_claim: Dict) -> bool:
        """Checks if the given token has expired or not"""
        now = datetime.datetime.now()
        if datetime.datetime.fromtimestamp(token_id_claim["exp"]) < now:
            return True
        return False

    def authenticate(self, **kwargs: Any) -> None:
        if not os.path.exists("outlook_token.json"):
            token = self._token(**kwargs)
            with open("outlook_token.json", "w") as token_json:
                token_json.write(json.dumps(token))

        elif os.path.exists("outlook_token.json"):
            with open("outlook_token.json", "r") as f:
                token = json.load(f)
                f.close()
            if self._token_expired(token["id_token_claims"]):
                token = self._token(**kwargs)
                with open("outlook_token.json", "w") as token_json:
                    token_json.write(json.dumps(token))
        self.service = requests_oauthlib.OAuth2Session(token=token)

    def get_latest_emails(self, **kwargs: Any) -> Sequence[IMessage]:
        if not self.service:
            raise ValueError("You need to authenticate Outlook API")
        messages = self.service.get(
            f"""
            https://graph.microsoft.com/v1.0/me/messages?$top=
            {kwargs.get('outlook_max_emails',10)}&$filter=isRead ne true"""
        ).json()
        messages = messages.get("value")
        messages_list = []
        for message in messages:
            raw_message_payload = self.get_email(outlook_id=message.get("id")).json()
            message = OutlookMessage.parse_message(raw_message_payload=raw_message_payload)
            if message:
                messages_list.append(message)
        return messages_list

    def get_email(self, **kwargs: Any) -> Any:
        return self.service.get(
            f"https://graph.microsoft.com/v1.0/me/messages/{kwargs.get('outlook_id')}"
        )



from __future__ import annotations

from abc import ABC
from typing import List, Union

from langchain.chains import LLMChain
from langchain.llms import HuggingFaceHub
from langchain.prompts import PromptTemplate

from inarrator.email.message import IMessage
from inarrator.summarizer.prompt import HUGGING_FACE_EMAIL_PROMPT


class HuggingFaceModel(ABC):
    def __init__(self, api_token: str, model_name: str):
        self.chain = self.load_chain(api_token, model_name)

    @classmethod
    def load_chain(self, api_token: str, model_name: str) -> LLMChain:
        return LLMChain(
            llm=HuggingFaceHub(huggingfacehub_api_token=api_token, repo_id=model_name),
            prompt=PromptTemplate(
                template=HUGGING_FACE_EMAIL_PROMPT, input_variables=["fro", "to", "subject", "body"]
            ),
        )

    def summarize(self, messages: Union[IMessage, List[IMessage]]) -> str:
        if isinstance(messages, list):
            raise ValueError(
                "HuggingFace Hub Model do not currently support summarizing multiple messages"
            )
        elif isinstance(messages, IMessage):
            return self.chain.predict(
                fro=messages.fro,
                to=messages.to,
                body=messages.body,
                subject=messages.subject,
            )
